# -*- coding: utf-8 -*-
#
# Geonum is a Python library for geographical calculations in 3D
# Copyright (C) 2017 Jonas Gliss (jonasgliss@gmail.com)
#
# This program is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License a
# published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""
Access and handling of topographic data
"""
import os

from geonum.exceptions import InvalidTopoMode
from geonum.topoaccessbase import SRTMAccess, Etopo1Access

class TopoDataAccess(object):
    """Factory class for accessing topographic data
    
    This is a high-level factory class which handles the access of topography
    data. Registered topographic datasets can be found in :attr:`REGISTERED` 
    
    Default access mode is SRTM (topographic dataset from NASA 
    `Shuttle Radar Topography Mission <https://www2.jpl.nasa.gov/srtm/>`__). 
    
    Example
    -------
    
    >>> acc = TopoDataAccess(mode='srtm')
    >>> topo_data = acc.get_data(lat0=5, lon0=30, lat1=15, lon1=40)
        
    Note
    ----
    
    For developers: the registered access  classes (such as 
    :class:`SRTMAccess`) should be based on (or follow the API of) template 
    base class :class:`TopoAccessBase`.
    
    
    Attributes
    ----------
    mode : str
        current access mode (string specifying which topographic dataset is
        supposed to be used for access). 
    local_path : str
        local path to etopo data (only relevant for etopo1 mode)
        
    Parameters
    ----------
    mode : str
        one of the supported data access types (cf. keys of :attr:`REGISTERED`)
    local_path : str
        local path to etopo data (only relevant for etopo1 mode)
    """
    #: supported access modes (topographic datasets)
    REGISTERED = dict(srtm      = SRTMAccess,
                      etopo1    = Etopo1Access)
    
    def __init__(self, mode="srtm", local_path=None):
        #mode and data access variables
        if not mode in self.REGISTERED:
            raise InvalidTopoMode(mode)
        self.mode = mode
    
        self.local_path = local_path
        
    @property
    def modes(self):
        """List of supported topographic datasets"""
        return list(self.REGISTERED.keys())
    
    @property
    def supported(self):
        """List of supported datasets (wrapper for :attr:`modes`)"""
        return self.modes
    
    def __deepcopy__(self, memo):
        return TopoDataAccess(self.mode, self.local_path)
    
    def get_data(self, lat0, lon0, lat1=None, lon1=None, 
                 mode=None, local_path=None, **access_opts):
        """Retrieve data from topography file
        
        Parameters
        ----------
        lat0 : float 
            start longitude for data extraction
        lon0 : float 
            start latitude for data extraction
        lat1 : float 
            stop longitude for data extraction (default: None). If None only 
            data around lon0, lat0 will be extracted.
        lon1 : float
            stop latitude for data extraction (default: None). 
            If None only data around lon0, lat0 will be extracted
        mode : str, optional
            mode specifying the topographic dataset that is supposed to be used
        local_path : str, optional
            local path where topography data is stored (can be dictionary or
            filepath, is passed to corresponding access class and handled 
            as implemented there)
        **access_opts
            additional access options that may be specific for the mode 
            specified (e.g. search_database in case of etopo1)
        
        
        Returns
        -------
        TopoData
            object containing the data
        
        Raises
        ------
        TopoAccessError
            if access fails
        
        """
        if mode is not None:
            if not mode in self.REGISTERED:
                raise InvalidTopoMode(mode)
            self.mode = mode
        if local_path is not None and os.path.exists(local_path):
            self.local_path=local_path
        access = self.REGISTERED[self.mode](local_path=self.local_path,
                                             **access_opts)
        return access.get_data(lat0, lon0, lat1, lon1)


# -*- coding: utf-8 -*-
"""Test environment for base.py module."""
from __future__ import (absolute_import, division)
import numpy.testing as npt
import pytest
from geonum.test.markers import srtm_avail
from geonum import TopoDataAccess

@pytest.fixture(scope='session')
@srtm_avail
def guallatiri_data():
    # Coordinates around Guallatiri volcano
    lat0 = -18.55
    lon0 = -69.2
    lat1 = -18.35  
    lon1 = -69.0
    
    acc = TopoDataAccess()
    
    return acc.get_data(lat0, lon0, lat1, lon1)

@srtm_avail    
def test_guallatiri_topo(guallatiri_data):
    data = guallatiri_data
    npt.assert_array_equal(data.shape, (242, 122))
    npt.assert_array_almost_equal([data.min, data.max, data.mean(), data.std()],
                                  [4084., 6057., 4690.692827, 393.816227])
    
    return data
    
if __name__ == '__main__':
    import matplotlib.pyplot as plt
    plt.close('all')
    
    data = guallatiri_data()
    
    data.plot_2d()
    
    data.plot_3d()
    
    test_guallatiri_topo(data)
    
    
    



# -*- coding: utf-8 -*-
"""Test environment for base.py module."""
from __future__ import (absolute_import, division)
import numpy.testing as npt
from geonum import GeoPoint, GeoVector3D

def test_GeoPoint():
    """Test basic arithmetic operations on GeoPoints."""
    p = GeoPoint(lat=0, lon=0, auto_topo_access=False)
    assert p.latitude == 0
    assert p.longitude == 0

def test_GeoVector3D():
    """Test basic arithmetic operations on GeoVectors."""
    v = GeoVector3D(dx=1, dy=1, dz=100)
    
    npt.assert_array_equal([v.dx, v.dy, v.dz], [1,1,100])
    npt.assert_array_almost_equal([v.elevation, 
                                   v.magnitude, 
                                   v.azimuth, 
                                   v.dist_hor], 
                                  [4.044691, 1.417745, 45., 1.414214])
def test_diffvector():
    p1 = GeoPoint(lat=37.751005, lon=14.993435, altitude=3264.0,
                  auto_topo_access=False)
    p2 = GeoPoint(37.765755,  15.016696, altitude=2820.0,
                  auto_topo_access=False)
    connection_vector = p2 - p1
    
    assert connection_vector.anchor is p1
    
    actual = [connection_vector.azimuth,
              connection_vector.elevation, 
              connection_vector.magnitude]
    
    
    npt.assert_allclose(actual=actual,
                        desired=[51.378677249653983,
                                 -9.6064174085658465,
                                 2.6606074796318557], rtol=1e-7)
    
if __name__ == '__main__':
    test_GeoPoint()
    test_GeoVector3D()
    test_diffvector()


# -*- coding: utf-8 -*-
#
# Geonum is a Python library for geographical calculations in 3D
# Copyright (C) 2017 Jonas Gliss (jonasgliss@gmail.com)
#
# This program is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License a
# published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""
Access and handling of topographic data
"""
import numpy as np
import os
from warnings import warn
# python 2 and 3 support see e.g.
# http://python-future.org/compatible_idioms.html#metaclasses
from six import with_metaclass

import abc
 
from geonum import NETCDF_AVAILABLE, SRTM_AVAILABLE
from geonum.exceptions import (TopoAccessError, SRTMNotCoveredError, 
                               InvalidTopoMode)
   
class TopoAccessBase(with_metaclass(abc.ABCMeta, object)):
    """Abstract base class for topgraphy file implementations
    
    Defines minimum interface for derived access classes of different 
    topographic datasets. 
    """
    #local_path = None
    #topo_id = None
    
    #: A coordinate for which data should be available
    _TESTLAT = 45
    _TESTLON = 15
    
    def __init__(self, local_path=None, check_access=True):
        self.local_path = local_path
        self.topo_id = None
        
        if check_access:
            self.check_access()
    
    @abc.abstractmethod
    def get_data(self, lat0, lon0, lat1=None, lon1=None):
        """Declaration of data access method 
        
        It is obligatory to implement this method into derived classes.
        
        Parameters
        ----------
        lat0 : float
            first latitude coordinate of topographic range (lower left coord)
        lon0 : float
            first longitude coordinate of topographic range (lower left coord)
        lat1 : int or float, optional
            second latitude coordinate of topographic range (upper right 
            coord). If None only data around lon0, lat0 will be extracted.
        lon1 : int or float, optional
            second longitude coordinate of topographic range (upper right 
            coord). If None only data around lon0, lat0 will be extracted.
        
        Returns
        -------
        TopoData
            instance of TopoData class
        """
        pass
    
    def check_access(self):
        """Check if topography data can be accessed"""
        from geonum.topodata import TopoData
        try: 
            d = self.get_data(self._TESTLAT, self._TESTLON)
            if not isinstance(d, TopoData):
                raise ValueError('Invalid return type, expected instance '
                                 'of TopoData class, got {}'.format(type(d)))
            return True
        except Exception as e:
            print('Could not access topodata: {}'.format(repr(e)))
        return False
            
    def _prep_borders(self, lat0, lon0, lat1, lon1):
        """Sort by longitudes and determines LL and TR coordinates
        
        Parameters
        ----------
        lat0 : float
            first latitude coordinate of topographic range (lower left coord)
        lon0 : float
            first longitude coordinate of topographic range (lower left coord)
        lat1 : float
            second latitude coordinate of topographic range (upper right coord)
        lon1 : float
            second longitude coordinate of topographic range (upper right coord)
        
        Returns
        -------
        tuple
            4-element tuple, containing:
                
            - float, smallest latitude (LL corner) 
            - float, smallest longitude (LL corner)
            - float, largest latitude (TR corner)
            - float, largest longitude (TR corner)
        """
        lats, lons = np.asarray([lat0, lat1]), np.asarray([lon0, lon1])
        return (np.nanmin(lats), np.nanmin(lons), 
                np.nanmax(lats), np.nanmax(lons))
        
    def _init_lons_lats(self, lats_all, lons_all, lat0, lon0, lat1=None, 
                        lon1=None):
        """Get all latitudes and longitudes on a topodata grid 
        
        Parameters
        ----------
        lats_all : ndarray 
            numpy array containing available latitudes of the accessed topo 
            dataset
        lons_all : ndarray
            numpy array containing available longitudes of the accessed topo 
            dataset
        lat0 : float
            first latitude coordinate of topographic range (lower left coord)
        lon0 : float
            first longitude coordinate of topographic range (lower left coord)
        lat1 : float, optional
            second latitude coordinate of topographic range (upper right coord)
        lon1 : float, optional
            second longitude coordinate of topographic range (upper right coord)

        Returns
        -------
        tuple
            2-element tuple, containing
            
            - ndarray, topodata latitudes overlapping with input range
            - ndarray, topodata longitudes overlapping with input range
        """
        if any([x is None for x in [lat1, lon1]]):
            lat1, lon1 = lat0, lon0
        if lon0 > lon1:
            lon0, lon1 = lon1, lon0
            lat0, lat1 = lat1, lat0
        #print lat0, lon0, lat1, lon1     
        #closest indices
        idx_lons = [np.argmin(abs(lons_all - lon0)), 
                    np.argmin(abs(lons_all - lon1))]
        idx_lats = [np.argmin(abs(lats_all - lat0)), 
                    np.argmin(abs(lats_all - lat1))]
        #Make sure that the retrieved indices actually INCLUDE the input ranges
        if idx_lons[0] == 0 and lons_all[0] > lon0:
            warn("Error: Lon0 smaller than range covered by file, using first"
                " available index in topodata..")
            lon0 = lons_all[0]
            idx_lons[0] = 0
        elif lons_all[idx_lons[0]] > lon0:
            idx_lons[0] -= 1
        if idx_lons[1] == len(lons_all) - 1 and lons_all[-1] < lon1:
            warn("Error: Lon1 larger than range covered by file, using last"
                " available index in topodata..")
            lon1 = lons_all[-1]
            idx_lons[1] = len(lons_all) - 1
        elif lons_all[idx_lons[1]] < lon1:
            idx_lons[1] += 1
        if idx_lats[0] == 0 and lats_all[0] > lat0:
            warn("Error: Lat0 smaller than range covered by file, using first"
                " available index in topodata..")
            lat0 = lats_all[0]
            idx_lats[0] = 0
        elif lats_all[idx_lats[0]] > lat0:
            idx_lats[0] -= 1
        if idx_lats[1] == len(lats_all) - 1 and lats_all[-1] < lat1:
            warn("Error: Lat1 larger than range covered by file, using last"
                " available index in topodata..")
            lat1 = lats_all[-1]
            idx_lats[1] = len(lats_all) - 1
        elif lats_all[idx_lats[1]] < lat1:
            idx_lats[1] += 1
        #make sure that no odd array lengths occur
        if not (idx_lats[1] - idx_lats[0] + 1) %2 == 0:
            #try append index at the end
            if not idx_lats[1] == len(lats_all) - 1:
                idx_lats[1] += 1
            elif not idx_lats[0] == 0:
                idx_lats[0] -= 1
            else:
                raise ValueError("Fatal error, odd length of latitude array")
        if not (idx_lons[1] - idx_lons[0] + 1) %2 == 0:
            #try append index at the end
            if not idx_lons[1] == len(lons_all) - 1:
                idx_lons[1] += 1
            elif not idx_lons[0] == 0:
                idx_lons[0] -= 1
            else:
                raise ValueError("Fatal error, odd length of longitude array")
        if idx_lats[0] > idx_lats[1]:
            return (lats_all[idx_lats[1] : idx_lats[0] + 1], 
                    lons_all[idx_lons[0] : idx_lons[1] + 1],
                    idx_lats, idx_lons)
        else:
            return (lats_all[idx_lats[0] : idx_lats[1] + 1],
                    lons_all[idx_lons[0] : idx_lons[1] + 1],
                    idx_lats, idx_lons)
       
class Etopo1Access(TopoAccessBase):
    """A class representing netCDF4 data access of Etopo1 data
    
    See `here <https://github.com/jgliss/geonum#supported-etopo1-files>`_ for
    instructions on the data access.
    
    Attributes
    ----------
    loader 
        data loader (:class:`netCDF4.Dataset`)
    local_path : str
        directory where Etopo1 data files are stored
    file_name :  str
        file name of etopo data file
        
    Parameters
    ----------    
    local_path : str 
        directory where Etopo1 data files are stored
    file_name :  str
        file name of etopo data file
    check_access : bool
        if True, then access to topography data is checked on init and an
        error is raised if no dataset can be accessed
    search_database : bool
        if True and topodata file :attr:`file_path` does not exist, then
        a valid topography file is searched in all paths that are specified
        in file `LOCAL_TOPO_PATHS.txt` that is shipped with this library and 
        can be found in installation subdirectory local_topo_data.
        
    Raises
    ------
    TopoAccessError
        if input arg `check_access` is True and if no supported data file 
        can be found
    
    """
    #: ID of dataset
    topo_id = "etopo1"
    
    #: filenames of supported topographic datasets in preferred order
    supported_topo_files = ["ETOPO1_Ice_g_gmt4.grd",
                            "ETOPO1_Bed_g_gmt4.grd"]
    def __init__(self, local_path=None, file_name=None, check_access=False,
                 search_database=True):
        
        if not NETCDF_AVAILABLE:
            raise ImportError("Etopo1Access class cannot be initiated. "
                                      "Please install netCDF4 library first")
        self._local_path = None
        self._file_name = None
        
        from netCDF4 import Dataset
        
        self.loader = Dataset
        
        self.local_path = local_path
        self.file_name = file_name
        
        if not os.path.exists(self.file_path) and search_database:
            self.search_topo_file_database()
        
        # check if file exists
        if check_access:
            if not os.path.exists(self.file_path):
                raise TopoAccessError('File {} could not be found in local '
                                      'topo directory: {}'.format(self.file_name, 
                                                       self.local_path))
            elif not self.check_access():
                raise TopoAccessError('Failed to extract topography data for '
                                      'Etopo dataset')
    
    @property
    def local_path(self):
        """Directory containing ETOPO1 gridded data files"""
        return self._local_path
    
    @local_path.setter
    def local_path(self, val):
        if val is None or not os.path.exists(val):
            from geonum import LOCAL_TOPO_PATH
            print('Invalid input for local_path, setting default: '
                  '{}'.format(LOCAL_TOPO_PATH))
            val = LOCAL_TOPO_PATH
        self._check_topo_path(val)
        self._local_path = val
    
    @property
    def file_name(self):
        """File name of topographic dataset used"""
        return self._file_name
    
    @file_name.setter
    def file_name(self, val):
        if not val in self.supported_topo_files:
            
            _val = self.supported_topo_files[0]
            print('Invalid input for file_name ({})), setting default: '
                  '{}'.format(val, _val))
            val = _val
        self._file_name = val
        
    @property
    def file_path(self):
        """Return full file path of current topography file"""
        return os.path.join(self.local_path, self.file_name)
    
    @file_path.setter
    def file_path(self, val):
        self.set_file_location(val)
    
    def _check_topo_path(self, path):
        """Check if path exists and if it is already included in database
        
        Parameters
        ----------
        path : str
            path to be checked
        """
        from geonum.helpers import check_and_add_topodir
        check_and_add_topodir(path)
            
    def _get_all_local_topo_paths(self):
        """Get all search paths for topography files"""
        from geonum.helpers import all_topodata_search_dirs
        return all_topodata_search_dirs()
        
    def _search_topo_file(self, path=None):
        """Checks if a valid etopo data file can be found in local folder
        
        Searches in ``self.local_path`` for any of the file names specified 
        in ``supported_topo_files``
        
        """
        if path is None:
            path = self.local_path
        print(("Searching valid topo file in folder: %s" %path))
        fnames = os.listdir(path)
        for name in fnames:
            if name in self.supported_topo_files:
                self.file_name = name
                self.local_path = path
                print(("Found match, setting current filepath: %s" 
                %self.file_path))
                return True
        return False
        
    def _find_supported_files(self):
        """Look for all supported files in ``self.local_path```and return list"""
        files = os.listdir(self.local_path)
        lst = []
        for name in files:
            if name in self.supported_topo_files:
                lst.append(name)
        return lst
    
    def search_topo_file_database(self):
        """Checks if a valid topo file can be found in database"""
        all_paths = self._get_all_local_topo_paths()
        for path in all_paths:
            if self._search_topo_file(path):
                return True
        return False
        
    def set_file_location(self, full_path):
        """Set the full file path of a topography data file
        
        Parameters
        ----------
        full_path : str
            full file path of topography file
            
        Raises
        ------
        TopoAccessError
            if filepath does not exist or if the provided file is not 
            supported by this interface.
        """
        if not os.path.exists(full_path):
            raise TopoAccessError('Input file location %s does not exist'
                                  .format(full_path))
        _dir = os.path.dirname(full_path)
        _f = os.path.basename(full_path)
        if not _f in self.supported_topo_files:
            raise TopoAccessError('Invalid topography data file name, please '
                                  'use either of the supported files from the '
                                  'Etopo1 data set: {}'
                                  .format(self.supported_topo_files))
        self.local_path = _dir
        self.file_name = _f
        
        if not os.path.basename(full_path) in self.supported_topo_files:
            raise TopoAccessError("Invalid topography data file, please use "
                "one of the supported files from the Etopo1 data set\n%s" 
                %self.supported_topo_files)
        self.local_path = os.path.dirname(full_path)
        self.file_name = os.path.basename(full_path)
    
    def get_data(self, lat0, lon0, lat1=None, lon1=None):
        """Retrieve data from topography file 
        
        Parameters
        ----------
        lat0 : float
            first latitude coordinate of topographic range (lower left coord)
        lon0 : float
            first longitude coordinate of topographic range (lower left coord)
        lat1 : int or float, optional
            second latitude coordinate of topographic range (upper right 
            coord). If None only data around lon0, lat0 will be extracted.
        lon1 : int or float, optional
            second longitude coordinate of topographic range (upper right 
            coord). If None only data around lon0, lat0 will be extracted.
        
        Returns
        -------
        TopoData
            instance of TopoData class
        """
        from geonum import TopoData
        etopo1 = self.loader(self.file_path)

        lons = etopo1.variables["x"][:]
        lats = etopo1.variables["y"][:]
        
        lats, lons, idx_lats, idx_lons = self._init_lons_lats(lats, lons, lat0,
                                                              lon0, lat1, lon1)

        vals = np.asarray(etopo1.variables["z"][idx_lats[0] : idx_lats[1] + 1,
                                                idx_lons[0] : idx_lons[1] + 1], 
                                                dtype = float)
        etopo1.close()

        return TopoData(lats, lons, vals, data_id=self.topo_id)
         
class SRTMAccess(TopoAccessBase):
    """Class for SRTM topographic data access
    
    Uses library `srtm.py <https://pypi.python.org/pypi/SRTM.py/0.3.1>`_ 
    for online access of data. 
    
    Note
    ----
    :mod:`srtm.py` downloads the topo data from `this source <http://
    dds.cr.usgs.gov/srtm/version2_1/>`_ and stores a copy of the unzipped data 
    files in the current cache directory found in home.
    
    Whenever data access is requested, the :mod:`srtm.py` checks if the file
    already exists on the local machine and if not downloads it online. The
    online access is rather slow, so do not be surprised, if things take a
    while when accessing a specific location for the first time. 
    
    **Deleting cached SRTM files**: 
        use :func:`geonum.topoaccess.delete_all_local_srtm_files`
      
    Parameters
    ----------
    check_access : bool
        check if data can be accessed on class initialisation
    **kwargs
        additional keyword arguments that are passed through (irrelevant for 
        this class but relevant for factory loader class 
        :class:`TopoDataAccess`, particularly :func:`set_mode` therein.
    """
    
    def __init__(self, check_access=False, **kwargs):
        """Class initialisation"""
        if not SRTM_AVAILABLE:
            raise ImportError("SRTM access class cannot be initiated. "
                                      "Please install srtm.py library first")
        import srtm
        self.loader = srtm
        self.topo_id = "srtm"
        
        if check_access:
            self.check_access()
        
    def _coordinate_covered(self, access_obj, lat, lon):
        """Checks if SRTM data is available for input coordinate
        
        Parameters
        ----------
        access_obj : GeoElevationData
            data access object from :mod:`srtm.py` module 
            (can be created calling ``srtm.get_data()``)
        lat : float
            latitude of point
        lon : float  
            longitude of point
        
        Returns
        -------
        bool
            True, if SRTM data is available for coordinate, else False.
        """
        if access_obj.get_file_name(lat, lon) is None:
            return False
        return True

        
    def get_data(self, lat0, lon0, lat1=None, lon1=None):
        """Load SRTM topographic subset for input range
        
        Parameters
        ----------
        lat0 : float
            first latitude coordinate of topographic range (lower left coord)
        lon0 : float
            first longitude coordinate of topographic range (lower left coord)
        lat1 : int or float, optional
            second latitude coordinate of topographic range (upper right 
            coord). If None only data around lon0, lat0 will be extracted.
        lon1 : int or float, optional
            second longitude coordinate of topographic range (upper right 
            coord). If None only data around lon0, lat0 will be extracted.
            
        Returns
        -------
        TopoData
            instance of TopoData class
        """
        from geonum import TopoData
        print("Retrieving SRTM data (this might take a while) ... ")
        # create GeoElevationData object for data access
        dat = self.loader.get_data()
        # check if second input point is specified and set equal first point if
        # not
        if any([x is None for x in [lat1, lon1]]):
            lat1, lon1 = lat0, lon0
        # Check if first point is covered by dataset
        if not self._coordinate_covered(dat, lat0, lon0):
            raise SRTMNotCoveredError('Point (lat={:.2f}, lon={:.2f}) not '
                                      'covered by SRTM'.format(lat0, lon0))
        # check if second point is covered by dataset
        if not self._coordinate_covered(dat, lat1, lon1):
            raise SRTMNotCoveredError('Endpoint coordinate (lat={:.2f}, '
                                      'lon={:.2f}) not covered by SRTM'
                                      .format(lat1, lon1))
        # prepare borders of covered lon / lat regime
        lat_ll, lon_ll, lat_tr,lon_tr = self._prep_borders(lat0, lon0,
                                                           lat1, lon1)
        # get SRTM file for lower left corner of regime
        f_ll = dat.get_file(lat_ll, lon_ll)
        # get SRTM file for top right corner of regime
        f_tr = dat.get_file(lat_tr, lon_tr)
        # create array of longitude values for regime
        lons_all = np.linspace(f_ll.longitude, f_tr.longitude + 1,
                            f_ll.square_side)
        # create array of latitude values for regime
        lats_all = np.linspace(f_ll.latitude, f_tr.latitude + 1,
                            f_ll.square_side)
        #prepare coordinates
        lats, lons, _, _= self._init_lons_lats(lats_all, lons_all,
                                               lat0, lon0, lat1, lon1)
        # Init data array 
        vals = np.ones((len(lats), len(lons))) * np.nan
        #loop over all coordinates and try access the elevation data
        for i in range(len(lats)):
            for j in range(len(lons)):
                #print "Lat: %s, Lon: %s" % (lats[i], lons[j])
                vals[i, j] = dat.get_elevation(lats[i], lons[j])

        return TopoData(lats, lons, vals, data_id=self.topo_id)

def delete_all_local_srtm_files():
    """Deletes all locally stored SRTM files"""
    if not SRTM_AVAILABLE:
        return
    import glob
    from srtm.main import FileHandler
    fh = FileHandler()
    for file in glob.glob('{}/*.hgt'.format(fh.get_srtm_dir())):
        print('Deleting SRTM data file at {}'.format(file))
        os.remove(file)
        
### Deprecated stuff
class TopoFile(TopoAccessBase):
    """Old name of :class:`TopoAccessBase`."""

    def __init__(self, *args, **kwargs):
        super(TopoFile, self).__init__(*args, **kwargs)
        warn(DeprecationWarning('Class TopoFile was renamed to TopoAccessBase'
                                'Please use the latter name'))


# -*- coding: utf-8 -*-
#
# Geonum is a Python library for geographical calculations in 3D
# Copyright (C) 2017 Jonas Gliss (jonasgliss@gmail.com)
#
# This program is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License a
# published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from LatLon23 import LatLon, GeoVector
    
from numpy import (radians, cos, sin, degrees, sqrt,tan, isnan, arctan2,
                   asarray, nanmin, nanmax)

from warnings import warn

class GeoPoint(LatLon):
    """The Geopoint object represents a location in the atmosphere 
    
    This object is in 3D and includes elevation information.
    
    Attributes
    ----------
    altitude : float
        altitude in m
    altitude_err : float
        uncertainty in altitude
    local_topo_path : str
        directory that is used to search for local topography if topodata is
        requested. Irrelevant for default topo
    Parameters
    ----------
    lat : float 
        latitude of point (decimal degrees)
    lon : float 
        longitude of point (decimal degrees)
    altitude : float
        elevation of point in m
    name : str
        name (ID) of this point
    topo_access_mode : str
        string specifying the current access mode for topographic data (in v1, 
        choose between srtm or etopo1)
    topo_path : str
        specify path where etopo1 data files are stored
    topo_data : TopoData 
        existing topographic dataset that can be assigned to this point (can 
        save time, e.g. for altitude access)
    """
    _ALTERR_DEFAULT = 99999999999.9
    def __init__(self, lat=0.0, lon=0.0, altitude=None, name="n/d",
                 topo_access_mode="srtm", topo_path=None, topo_data=None,
                 auto_topo_access=True):
        
        LatLon.__init__(self, float(lat), float(lon), name)
        self.altitude = altitude #altitude in m
        self.altitude_err = 0.0
        
        self.topo_access_mode = topo_access_mode
        self.local_topo_path = topo_path
        
        self.topo_data = None 
        
        if topo_data is not None:
            self.set_topo_data(topo_data)
        
        if self.altitude is None or isnan(self.altitude):
            if auto_topo_access:
                self.get_altitude()
            else:
                self.altitude = 0.0
                self.altitude_err = self._ALTERR_DEFAULT
    
    @property
    def longitude(self):
        """Longitude coordinate in decimal degrees"""
        return self.lon.decimal_degree

    @property
    def latitude(self):
        """Latitude coordinate in decimal degrees"""
        return self.lat.decimal_degree
    
    def offset(self, azimuth, dist_hor, dist_vert=0.0, ellipse="WGS84", 
               **kwargs):
        """Returns new GeoPoint at offset position
        
        Parameters
        -----------
        azimuth : float
            azimuth direction angle (in decimal degrees, e.g. 90 for E 
            direction, -90 or 270 for west)
        dist_hor : float 
            horizontal offset to this point in km 
        dist_vert : float 
            vertical offset to this point in m (positive if point is higher, 
            negative, if it is lower)
        ellipse : float 
            geodetic system (ellipsoid), default is "WGS84", i.e. 
            `World Geodetic System <https://confluence.qps.nl/pages/view page.
            action?pageId=29855173>`__
        **kwargs : 
            additional keyword arguments passed to init of new 
            :class:`GeoPoint` object
        
        Returns
        -------
        GeoPoint
            new location at offset position
        """
        p = LatLon(self.lat.decimal_degree, self.lon.decimal_degree)
        p1 = p.offset(azimuth, dist_hor, ellipse)
        return GeoPoint(p1.lat.decimal_degree, p1.lon.decimal_degree,
                        self.altitude + dist_vert, **kwargs)
    
    def set_topo_data(self, topo_data):
        """Assign topographic data to this point
        
        If input is valid, the topographic data set will be stored in 
        class attribute topo_data.
        
        Parameters
        ----------
        topo_data : TopoData
            topography dataset
        """
        from geonum.topodata import TopoData
        if not isinstance(topo_data, TopoData):
            return
        elif not topo_data.includes_coordinate(self.latitude, self.longitude):
            raise TypeError("Could not assign topodata to GeoPoint %s: "
             "topodata does not cover the coordinates" %self.name)
        self.topo_data = topo_data
    
    def range_borders(self, *points):
        """Get geographical borders (lower left, top right) of this point
        
        Additional points can be included as non keyword arguments. The
        range borders are then determined under consideration of all
        specified points. If no additional points are specified,
        the range corners will be set corresponding to 1km diagonal 
        distance to this point.
        
        Parameters
        ----------
        *points 
            additional :class:`GeoPoint` objects to be  considered
        
        Returns
        -------
        2-element tuple, containing:
            
            - :class:`GeoPoint`, lower left corner of regime
            - :class:`GeoPoint`, top right corner of regime
        
        """
        lats, lons = [self.latitude], [self.longitude]
        #retrieve all latitudes and longitudes
        for p in points:
            if isinstance(p, GeoPoint):
                lats.append(p.latitude)
                lons.append(p.longitude)
        lats, lons = asarray(lats), asarray(lons)
        if not len(lats) > 0:
            #print "Borders could not be initiated, no objects found..."
            return False
        lat_ll, lon_ll, lat_tr , lon_tr = (nanmin(lats), nanmin(lons), 
                                           nanmax(lats), nanmax(lons))
        pll, ptr = GeoPoint(lat_ll, lon_ll, 0.0), GeoPoint(lat_tr, lon_tr, 0.0)
        extend_km = (pll - ptr).magnitude * 0.1
        if extend_km == 0:
            extend_km = 1
        ll = pll.offset(azimuth=-135, dist_hor=float(extend_km), name="ll")
        tr = ptr.offset(azimuth=45, dist_hor=float(extend_km), name="tr")
        return  ll, tr        
        
    def get_topo_data(self, geo_point=None, azimuth=None, dist_hor=None,
                      lon1=None, lat1=None):
        """Retrieve topographic data grid spanned by this and another point
        
        The second coordinate can be specified in several different ways 
        dependent on which of the input arguments are used. First checks if the
        currently loaded topo data (``self.topo_data``) covers the range 
        spanned by the 2 points. If this is the case, the loaded data will be 
        used and nothing reloaded. If not, data will attempted to be loaded 
        from the current :class:`TopoDataAccess` object within the lon lat 
        range covered by the 2 points (set using :func:`range_borders`)
        
        Note
        ----
        The following input combinations work (and are preferentially 
        processed in the specified list order if multiple input is given):
        
            1. specify endpoint using geo_point
            #. specify endpoint using azimuth and dist_hor
            #. specify endpoint using lon1 and lat1
        
        Parameters
        ----------
        geo_point : GeoPoint 
            another geo_point,topo data will be retrieved between this point 
            and destination point
        azimuth : float 
            azimuth angle of heading in decimal degrees (to be used with 
            dist_hor)
        dist_hor : float 
            horizontal distance from this point in km (to be used with azimuth)
        lon1 : float
            longitude of destination point, topo data will be retrieved between 
            this point and destination point (to be used with lat1)
        lat1 : float 
            latitude of destination point, topo data will be retrieved between 
            this point and destination point (to be used with lon1)
        
        Returns
        -------
        2-element tuple, containing:
            
            - :class:`TopoData`, the topographic data
            - :class:`GeoPoint`, the second coordinate
        """
        pf = self
        if isinstance(geo_point, GeoPoint):
            pf = geo_point
        elif all(isinstance(x, (int, float)) for x in [dist_hor, azimuth]):
            pf = self.offset(azimuth, dist_hor)
        elif all(isinstance(x, (int, float)) for x in [lon1, lat1]):
            pf = GeoPoint(lat1, lon1)
        
        ll, tr = self.range_borders(self, pf)
        lat0, lon0 = ll.latitude, ll.longitude
        lat1, lon1 = tr.latitude, tr.longitude
        if not (self.check_topo(lat1, lon1) and self.check_topo(lat0, lon0)):
            print(("Topo data not avaialable in geo point %s, loading topodata"
                        %self.name))
            self._load_topo_data(lat0, lon0, lat1, lon1)
        else:
            print ("Topo data avaialable in geo point %s")
        return self.topo_data, pf
    
    def check_topo(self, lat1=None, lon1=None):
        """Check if topography is available between this point and another
        
        :param float lat1: latitude of end point, if None, only the 
            coordinates of this object will be considered
        :param float lon1: longitude of end point, if None, only the 
            coordinates of this object will be considered
        :returns: - bool, True if topo data is loaded, False if not
        """
        if any([x == None for x in [lat1, lon1]]):
            lat1, lon1 = self.latitude, self.longitude
        if self.topo_data is None:
            return False
        if not (self.topo_data.includes_coordinate(self.latitude, 
                                                   self.longitude) 
            and self.topo_data.includes_coordinate(lat1, lon1)):
            return False
        return True
        
    def get_elevation_profile(self, geo_point=None, azimuth=None, 
                              dist_hor=None, lon1=None, lat1=None, 
                              resolution=5., **mapping_opts):
        """Estimates the elevation profile for a given viewing direction up
        to a given distance from the coordinates of this point. For input 
        possibilities see docs of :func:`get_topo_data`.
        
        :param GeoPoint geo_point: another geo_point,topo data will be 
            retrieved between this point and destination point
        :param float azimuth: azimuth angle of heading in decimal degrees 
            (to be used with dist_hor)
        :param float dist_hor: horizontal distance from this point in km 
            (to be used with azimuth)
        :param float lon1: longitude of destination point, topo data will 
            be retrieved between this point and destination point (to be 
            used with lat1)
        :param float lat1: latitude of destination point, topo data will be 
            retrieved between this point and destination point 
            (to be used with lon1)
        :param float resolution: desired topo grid resolution in m. 
            1D interpolation of the elevation profile is performed is 
            applicable.
        :returns: 
            - :class:`ElevationProfile`, the profile object
          
        The following input combinations work (and are preferentially processed
        in the specified list order if multiple input is given):
        
            1. specify endpoint using geo_point
            #. specify endpoint using azimuth and dist_hor
            #. specify endpoint using lon1 and lat1
            
        
        """
        from geonum.elevationprofile import ElevationProfile
        data, pf = self.get_topo_data(geo_point, azimuth, dist_hor, lon1, 
                                      lat1)
        return ElevationProfile(data, self, pf, resolution=resolution,
                                **mapping_opts)
        
    def get_altitude(self):
        """Estimate the altitude (+/- uncertainty) of this point 
        
        The estimatation is done by retrieving a 2x2 grid of topography
        data enclosing this point. The altitude value is estimated using the 
        topo data tile closest to the coordinates of this point. The uncertainty
        is estimated conservatively using min/max difference of data in the 2x2
        grid.
        
        :returns: 
            - altitude (0 if retrieval failed)
            - altitude error (99999999999 if retrieval failes)
        :rtype:
            - int, float
        
        """
#==============================================================================
#         print ("Try retrieving altitude for point %s (lat, lon): %s, %s" 
#                                 %(self.name, self.latitude, self.longitude))
#==============================================================================
        try:
            #the data access class already catches access failure and in case,
            #tries to  access etopo1 data
            data = self._topo_access.get_data(self.latitude,
                                              self.longitude)
            z = data(self.latitude, self.longitude)# / 1000.
            z_err = (data.data.max() - data.data.min()) #/ 1000.'
#            print "Retrieved altitude: %s +/- %s m" %(z, z_err)
        except Exception as e:
            warn('Altitude of {} could not be retrieved...\n'
                 'Setting altitude to 0.0 m\n'
                 'Error: {}'.format(repr(self), repr(e)))
            z, z_err = 0.0, self._ALTERR_DEFAULT
    
        self.altitude, self.altitude_err = z, z_err
        
        return (z, z_err)
    
    """HELPERS, IO stuff, etc"""
    def update_topo_access(self, mode, local_path=None):
        """Update topo access mode
        
        :param str mode: valid access mode (e.g. "srtm", "etopo1")        
        :param str local_path (None): path for etopo1 access (if None, uses 
            the one which is currently set in ``self._topo_access``)
        """
        if local_path is not None:
            self.local_topo_path = local_path
        self.topo_access_mode = mode
        
        
    """Plotting etc..."""
    def plot_2d(self, map, add_name=False, dist_text=0.5, angle_text=-45, 
                **kwargs):
        """Plot this point into existing 2D basemap
                
        :param map basemap: Basemap object (drawn in an Axes3D object)
        :param bool add_name: add the name of this GeoPoint in the map
        :param float dist_text: distance of text annotation from point
        :param float angle_text: angle of text displacement
        :param kwargs: additional keyword arguments passed to matplotlib 
            plot function
        """
        if not "marker" in kwargs:
            kwargs["marker"] = "^"
        if not any([x in kwargs for x in ["c", "color"]]):
            kwargs["c"] = "lime"
        map.draw_geo_point_2d(self, **kwargs)
        if add_name:
            map.write_point_name_2d(self, dist_text, angle_text)
                                                    
    def plot_3d(self, map, add_name=False, dz_text=0.0, **kwargs):
        """Plot this point into existing 3D basemap
                
        :param map basemap: Basemap object (drawn in an Axes3D object)
        :param bool add_name: add the name of this GeoPoint in the map
        :param float dz_text: altitude offset of text(to point)
        :param kwargs: additional keyword arguments passed to matplotlib 
            plot function
        """
        if not "marker" in kwargs:
            kwargs["marker"] = "^"
        if not any([x in kwargs for x in ["c", "color"]]):
            kwargs["c"] = "lime"
        map.draw_geo_point_3d(self, **kwargs)
        x0, y0 = list(map(self.lon.decimal_degree, self.lat.decimal_degree))
        if add_name and self.name is not None:
            try:
                fs = kwargs["fontSize"]
            except:
                fs = 12
            #zt=self.altitude*1000. + dz_text
            zt = self.altitude + dz_text
            map.draw_text_3d(self.longitude, self.latitude, zt, self.name,
                             color="k", fontsize=fs)
            
    @property
    def _topo_access(self):
        """Topography data access class"""
        from geonum.topodataaccess import TopoDataAccess
        return TopoDataAccess(mode=self.topo_access_mode,
                              local_path=self.local_topo_path)
        
    def _load_topo_data(self, lat0, lon0, lat1, lon1):
        """Load and update topo data
        
        :param float lon0: start longitude
        :param float lat0: start latitude
        :param float lon1: stop longitude 
        :param float lat1: stop latitude
        :return:
            - :class:`TopoData` object with loaded data (it is also stored in
                ``self.topo_data``)
        """
        print("Trying to load topography data for {}".format(self))
        self.topo_data = self._topo_access.get_data(lat0, lon0, lat1, lon1)
        return self.topo_data
        
    def _sub_geo_vector_2d(self, other):
        """Subtract another geo vector
        
        Called when subtracting a GeoVector object from self (adapted from
        `LatLon` object,  only return type was changed from LatLon 
        object to GeoPoint object)
        
        :param (GeoVector, GeoVector3D) other: vector to be subtracted
        :return: :class:`GeoPoint` at difference position
        """
        heading, distance = other()
        heading = (heading + 180) % 360 # Flip heading
        p = GeoPoint(self.lat, self.lon, self.altitude)
        
        return p.offset(heading, distance)
    
    def _add_geo_vector_2d(self, other):
        """Subtract another geo vector
        
        Add a geo vector (adapted from `LatLon` object,  only return 
        type was changed from LatLon object to GeoPoint object)
        
        :param GeoVector other: Geovector added to this 
        
        """
        azimuth, dist_hor = other()
        p = GeoPoint(self.lat, self.lon, self.altitude)
        return p.offset(azimuth, dist_hor, 0.0)
        
    def _sub_geo_vector_3d(self, other):
        """Called when subtracting a GeoVector3D object from self"""
        azimuth, dist_hor, dz = other()
        azimuth = (azimuth + 180) % 360 # Flip heading
        p = GeoPoint(self.lat, self.lon, self.altitude) # Copy current position
        return p.offset(azimuth, dist_hor, -dz) # Offset position by GeoVector
    
    def _add_geo_vector_3d(self, other):
        """Add a 3d geo vector object
        
        :param GeoVector3D other: Geovector which is added to self
        """
        azimuth, dist_hor, dz = other()
        p = GeoPoint(self.lat, self.lon, self.altitude)
        return p.offset(azimuth, dist_hor, dz)
        
    def _sub_latlon(self, other):
        """Called when subtracting a LatLon object from this point"""
        from geonum import GeoVector3D
        inv = self._pyproj_inv(other)
        heading = inv['heading_reverse']
        distance = inv['distance']
        name=str(other.name + "->" + self.name)
        return GeoVector3D(dz=0.0, azimuth=heading, dist_hor=distance,
                           anchor=other, name=name)
    
    def _sub_geo_point(self, other):
        """Called when subtracting a LatLon object from self"""
        from geonum import GeoVector3D
        inv = self._pyproj_inv(other)
        heading = inv['heading_reverse']
        distance = inv['distance']
        name = str(other.name + "->" + self.name)
        dz = self.altitude - other.altitude
        return GeoVector3D(dz=dz, azimuth=heading, dist_hor=distance,
                           anchor=other, name=name)
        
    def __eq__(self, other):
        """Checks equality of this point with another
        
        :param other: another :class:`GeoPoint` object
        """
        if (self.lat == other.lat and self.lon == other.lon 
            and self.altitude == other.altitude):
            return True
        return False
        
    def __add__(self, other):
        """Add a geo vector to this point
        
        :param (GeoVector, GeoVector3D) other: a vector
        """
        object_operator = {'GeoVector'      :   self._add_geo_vector_2d,
                           'GeoVector3D'    :   self._add_geo_vector_3d}
        return object_operator[other.type()](other)

    def __sub__(self, other):
        """Subtraction"""
        object_operator = {'GeoVector'      :   self._sub_geo_vector_2d,
                           'GeoVector3D'    :   self._sub_geo_vector_3d,
                           'LatLon'         :   self._sub_latlon,
                           'GeoPoint'       :   self._sub_geo_point}
        return object_operator[other.type()](other)
    
    def __str__(self):
        """String formatting"""
        return ("GeoPoint %s\nLat: %s, Lon: %s, Alt: %s m\n" 
                %(self.name, self.latitude, self.longitude, self.altitude))
    
    def __repr__(self):
        """Obj. representation"""
        return ("%s, %s, Alt. %s m" 
                %(self.lat.__repr__(), self.lon.__repr__(), self.altitude))
    
    def __complex__(self):
        """Complex representation of lat and lon"""
        return self.complex()
    
    def type(self):
        """Object type identifier
        
        :returns: str, the string identifier        
        """
        return 'GeoPoint'



# -*- coding: utf-8 -*-
"""Test environment for base.py module."""
from __future__ import (absolute_import, division)
import numpy.testing as npt
import pytest
import numpy as np
from geonum import SRTM_AVAILABLE
from geonum import GeoPoint

@pytest.mark.skipif(SRTM_AVAILABLE==False,
                   reason='Skipping Elevation profile tests. srtm.py library is '
                   'not installed')
    
def test_ElevationProfile():
    
    lat_taranaki = -39.296571
    lon_observer = 173.9224
    
    obs = GeoPoint(lat_taranaki, lon_observer)
    
    npt.assert_almost_equal(obs.altitude, 324)
    
    prof = obs.get_elevation_profile(azimuth=90, dist_hor=50)
    
    prof_nans = obs.get_elevation_profile(azimuth=45, dist_hor=50)
    prof2 = obs.get_elevation_profile(azimuth=45, dist_hor=50,
                                      order=1)
    
    npt.assert_array_equal([len(prof.dists), len(prof.profile),
                            np.sum(np.isnan(prof_nans.profile))],
                            [10092,10092,10263])
    npt.assert_array_almost_equal([prof.profile.max(), prof.profile.min(), 
                                   prof.profile.mean(), prof.profile.std(),
                                   prof.dists.max(),
                                   prof2.profile.max(), prof2.profile.min()],
                                  [2482.598405, 176.008484, 583.012202,  
                                   486.57815, 50.033677, 1005.208932, 0.])
    
if __name__ == '__main__':
    import matplotlib.pyplot as plt
    
    plt.close('all')
    test_ElevationProfile()
    
# =============================================================================
#     prof.line.plot_line_on_grid(prof.topo_data.data)
#     
#     
#     d = prof.topo_data.data
#     
#     # all NaN
#     
#     z1 = prof.line.get_line_profile(d, prefilter=False)
#     z2 = prof.line.get_line_profile(d, order=1, prefilter=True)
#     z3 = prof.line.get_line_profile(d, order=1, prefilter=False)
#     
#     z0 = prof.line.get_line_profile(d)
#     
#     
#     d[np.isnan(d)] = 0
#     
#     zi = prof.line.get_line_profile(d)
#     
#     fig = plt.figure(figsize=(16, 8))
#     
#     
#     plt.plot(z1, label='prefilter=False')
#     plt.plot(z2, label='order=1, prefilter=True')
#     plt.plot(z3, ' x', label='order=1')
#     plt.plot(zi, label='default, after NaNs -> 0')
#     ax = fig.axes[0]
#     ax.set_yscale("log", nonposy='clip')
#     ax.set_ylim([0.1, 3000])
#     
#     #
#     
#     plt.legend()
# =============================================================================



# -*- coding: utf-8 -*-
#
# Geonum is a Python library for geographical calculations in 3D
# Copyright (C) 2017 Jonas Gliss (jonasgliss@gmail.com)
#
# This program is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License a
# published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import numpy as np
from geonum import (CV2_AVAILABLE, LATLON_AVAILABLE)
if LATLON_AVAILABLE:
    from LatLon23 import LatLon
   
class TopoData(object):
    """Data class for topography data
    
    This object represents topographic data in a certain latitude and longitude 
    range, specified by the corner coordinates of the range 
    (:attr:`lon0, lat0, lon1, lat1`). It may be used for 2D and 3D plotting of 
    the topographic data (cf. :func:`plot_2d`, :func:`plot_3d`) or for further
    analysis such as coordinate based altitude retrievals 
    (:func:`get_altitude`), change of the grid resolution (cf. 
    :func:`increase_grid_resolution`) or for customised analysis by directly
    using the numpy data array containing the topographic altitudes 
    (:attr:`data`), together with the corresponding dimension coordinates 
    which can be accessed via attributes :attr:`latitude`and :attr:`longitude`.
    
    To create an instance of this class, you may use the class 
    :class:`TopoDataAccess`.
    
    Note
    ----
    This object does not provide topography data access. For data access, 
    please use :class:`TopoDataAccess` which returns :class:`TopoData` 
    objects. This object is intended for storage and post processing of
    topographic data 
        
    Parameters
    ----------
    lats : ndarray 
        numpy array with latitude coordinates of the topographic dataset.
        Accessible via :attr:`latitude`.
    lons : ndarray 
        numpy array with longitude coordinates of the topographic dataset.
        Accessible via :attr:`longitude`.
    data : ndarray
        2D numpy array containing elevation values.
    data_id : str
        ID of this data set.
    repl_nan_minval : bool 
        coordinates containing NaN values are replaced with the minimum 
        altitude in the range.
    """
    def __init__(self, lats, lons, data, data_id="", repl_nan_minval=False):
        self.data_id = data_id #: ID of topodata file

        self.lats = lats #
        self.lons = lons #asarray(lons)
        
        if repl_nan_minval:
            self.replace_nans()
        
        self.data = data
        
    @property
    def latitude(self):
        """Wrapper for :attr:`lats`"""
        return self.lats
    
    @property
    def longitude(self):
        """Wrapper for :attr:`lats`"""
        return self.lats
    
    def replace_nans(self, fillval=None):
        """Replace NaNs in topographic data with a fill value
        
        Parameters
        ----------
        fillval : float
            value that is assigned to all coordinates that include NaNs
        """
        if fillval is None:
            fillval = np.nanmin(self.data)
        self.data[np.isnan(self.data)] = fillval
        
    @property
    def lon0(self):
        """Returns first longitude (i.e. ``self.lons[0]``)"""
        return self.lons[0]
        
    @property
    def lat0(self):
        """Returns first latitude (i.e. ``self.lats[0]``)"""
        return self.lats[0]
        
    @property
    def lon1(self):
        """Returns last longitude (i.e. ``self.lons[-1]``)"""
        return self.lons[-1]
        
    @property
    def lat1(self):
        """Returns last latitude (i.e. ``self.lats[-1]``)"""
        return self.lats[-1]
        
    @property
    def max(self):
        """Returns maximum altitude of topo data"""
        return np.nanmax(self.data)
        
    @property
    def min(self):
        """Returns minimum altitude of topo data"""
        return np.nanmin(self.data)
    
    @property
    def shape(self):
        """Shape of 2D numpy array that contains topography"""
        return self.data.shape
    
    @property
    def alt_range(self):
        """Returns covered altitude range"""
        return self.max - self.min
        
    @property
    def resolution(self):
        """Returns tuple (lat, lon) of the current grid resolution in km
        
        Note
        ----
        The resolution is determined at the center of this grid
        """
        if not LATLON_AVAILABLE:
            raise ImportError('Feature disabled: Neither LatLon nor '
                                      'LatLon23 are installed')
        x_lon, x_lat = int(len(self.lons) / 2), int(len(self.lats) / 2)
        p0 = LatLon(self.lats[x_lat], self.lons[x_lon])
        r_lon = (p0 - LatLon(self.lats[x_lat], self.lons[x_lon + 1])).magnitude
        r_lat = (p0 - LatLon(self.lats[x_lat + 1], self.lons[x_lon])).magnitude
        return (r_lat, r_lon)
    
    def mean(self):
        """Mean elevation (ignores NaNs in data)"""
        return np.nanmean(self.data)  
    
    def std(self):
        """Standard deviation of topographic dataset"""
        return np.nanstd(self.data)
        
    def increase_grid_resolution(self, res=0.2, polyorder=2):
        """Gaussian pyramide based upscaling of topographic grid
        
        This function checks the current topographic resolution in the center
        of the grid. Based on this, an upsacling factor is determined and the
        :func:`cv2.pyrUp` is used to increase the resolution using
        interpolation. Note, that this does not increase the actual resolution
        of the topographic data grid.
        
        Note
        ----
        Requires that opencv library is installed.
        
        Parameters
        ----------
        res : int or float, optional
            desired grid resolution in km (default: 0.2)
        polyorder : int, optional 
            order of polynomial used for interpolation (default: 2)
        
        Returns
        -------
        TopoData
            new object with desired grid resolution
        
        Raises
        ------
        ImportError
            if opencv is not installed.
        """
        if not CV2_AVAILABLE or not LATLON_AVAILABLE:
            raise ImportError('Feature disabled: Require opencv and '
                                      'LatLon (or LatLon23) library to change '
                                      'grid resolution ')
        from cv2 import pyrUp
        lons = self.lons
        lats = self.lats
        vals = self.data
        if not all(np.mod(x, 2) == 0 for x in [len(lons), len(lats)]):
            print ("Fatal error, odd array size detected, no upscaling possible."
                " Return current data dict")
            return False
            
        c_lon = len(lons) / 2 - 1 #center longitude index
        c_lat = len(lats) / 2 - 1 #center latitude index
        p1 = LatLon(lats[c_lat], lons[c_lon]) 
        p2 = LatLon(lats[c_lat + 1], lons[c_lon + 1])
        dist = (p2 - p1).magnitude #distance between 2 points on grid
        res_fac = dist / res #factor for increasing the spatial resolution
        if res_fac <= 1:
            print(("No interpolation of topodata necessary: topo raw data "
                "already has desired resolution...\nCurrent resolution: %s" 
                + str(dist) + "km\nDesired resolution: %s km" %(dist, res)))
            return self
        fac = int(np.ceil(np.log2(res_fac))) #the corresponding up-factor for the scale space
        print(("Increasing spatial topography resolution by factor %s" %(2**fac)))
        for k in range(fac):
            vals = pyrUp(vals)
        p_lons = np.poly1d(np.polyfit(np.arange(len(lons)), lons, polyorder))
        p_lats = np.poly1d(np.polyfit(np.arange(len(lats)), lats, polyorder))
        lons_new = p_lons(np.linspace(0, len(lons) - 1, vals.shape[1]))
        lats_new = p_lats(np.linspace(0, len(lats) - 1, vals.shape[0]))
        return TopoData(lats_new, lons_new, vals, self.data_id + "_interp")

    @property
    def delta_lon(self):
        """Longitude range of data (in decimal degrees)"""
        return self.lons[-1] - self.lons[0]
        
    @property 
    def delta_lat(self):
        """Latitude range of data (in decimal degrees)"""
        return self.lats[-1] - self.lats[0]
        
    @property
    def center_coordinates(self):
        """Tuple (lat, lon) with center coordinates of data"""
        return (self.lats[0] + self.delta_lat / 2.,
                self.lons[0] + self.delta_lon / 2.)
    
    def plot(self, plot3d=True, draw_coastlines=False, 
             draw_mapscale=False, **kwargs):
        from geonum.mapping import Map
        if not "projection" in kwargs:
            kwargs["projection"] = "lcc"
        if not "llcrnrlon" in kwargs:
            kwargs["llcrnrlat"] = self.lat0
            kwargs["llcrnrlon"] = self.lon0
            kwargs["urcrnrlat"] = self.lat1
            kwargs["urcrnrlon"] = self.lon1
        
            kwargs["lat_0"] = self.lat0 + (self.lat1-self.lat0)/2
            kwargs["lon_0"] = self.lon0 + (self.lon1-self.lon0)/2
        
        m = Map(**kwargs)
        m.set_topo_data(self)
        
        if plot3d:
            if draw_coastlines:
                raise NotImplementedError('3D plotting and coastline drawing is '
                                          'not supported yet')
            elif draw_mapscale:
                raise NotImplementedError('3D plotting and mapscale drawing is '
                                          'not supported yet')
            m.draw_topo_3d()
        else:  
            if draw_coastlines:
                m.draw_coastlines()
            m.draw_coordinates()
            if draw_mapscale:
                m.draw_mapscale_auto()
        return m
    
    def plot_2d(self, ax=None):
        """Plot 2D basemap of topodata"""
        from geonum.mapping import Map
        latc, lonc = self.center_coordinates
        m = Map(self.lon0, self.lat0, self.lon1, self.lat1,
                projection="merc", lat_0=latc, lon_0=lonc, ax=ax)
        m.set_topo_data(self)
        m.draw_topo(insert_colorbar=True)
        m.draw_topo_contour() 
        m.draw_coordinates()
        m.drawcoastlines()
        return m
        
    def plot_3d(self, ax=None):
        """Creates 3D surface plot of data
        
        Parameters
        ----------
        ax
            instance of matplotlib Axes3D object
        
        Returns
        -------
        geonum.mapping.Map
            plotted map object.
            
        """
        from geonum.mapping import Map
        latc, lonc = self.center_coordinates
        m = Map(self.lon0, self.lat0, self.lon1, self.lat1, 
                projection="merc", lat_0=latc, lon_0=lonc, ax=ax)
        m.set_topo_data(self)
        m.draw_topo_3d()
        return m
        
    def includes_coordinate(self, lat=None, lon=None):
        """Checks if input coordinates are covered by this dataset"""
        if not (self.lat0 <= lat <= self.lat1 and 
                self.lon0 <= lon <= self.lon1):
            print("Input coordinates out of topo data range...")
            return False
        return True
    
    def closest_index(self, lat, lon):
        """Finds closest index to input coordinate
        
        Parameters
        ----------
        lat : float
            latitude coordinate in decimal degrees
        lon : float 
            longitude coordinate in decimal degrees
        
        Returns
        -------
        tuple
            2-element tuple containing closest index of lat and lon arrays to
            to input index
        
        Raises
        ------
        ValueError
            if input coordinate is not included in this dataset
        """
        if not self.includes_coordinate(lat, lon):
            raise ValueError("Input values out of range...")
        
        idx_lat = np.argmin(abs(self.lats - lat))
        idx_lon = np.argmin(abs(self.lons - lon))
        return (idx_lat, idx_lon)
        
    def get_altitude(self, lat, lon):
        """Get altitude value at input coordinate
        
        Parameters
        ----------
        lat : int or float 
            latitude of coordinate
        lon : int or float 
            longitude of coordinate
            
        Returns
        -------
        float
            altitude at input coordinate
        
        Raises
        ------
        ValueError
            if retrieved altitude value is NaN
        """
        idx_lat, idx_lon = self.closest_index(lat, lon)
        dat = self.data[idx_lat, idx_lon]
        if np.isnan(dat):
            raise ValueError("Invalid value encountered in topodata...")
        return dat
        
    def __call__(self, lat, lon):
        """Get altitude value at input position"""
        try:
            return self.get_altitude(lat, lon)
        except ValueError as e:
            raise e


# -*- coding: utf-8 -*-
#
# Geonum is a Python library for geographical calculations in 3D
# Copyright (C) 2017 Jonas Gliss (jonasgliss@gmail.com)
#
# This program is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License a
# published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""
Processing module of geonum library
"""
import numpy as np
from scipy.ndimage import map_coordinates

class LineOnGrid(object):
    """A class representing a line on a discrete grid"""
    def __init__(self, x0, y0, x1, y1, id=""):
        """Class initialisation
        
        :param int x0: start x coordinate
        :param int y0: start y coordinate
        :param int x1: stop x coordinate
        :param int y1: stop y coordinate
        :param str id: string for identification (optional)
        """
        self.id = id
        self.start  = [x0,y0]
        self.stop   = [x1,y1]
        
        self.profile_coords = None
        
        #self.check_coordinates()
        self._det_normal_vec()
        self.prepare_profile_coordinates()
    
    def check_coordinates(self):
        """Check if coordinates are in the right order and swap, if not"""
        if any([x < 0 for x in self.start]):
            raise ValueError("Invalid value encountered, coords must not be"
                " smaller than 0")
        if any([x < 0 for x in self.stop]):
            raise ValueError("Invalid value encountered, coords must not be"
                " smaller than 0")
        if self.start[0] > self.stop[0]:
            print("X coordinate of Start point is larger than X of stop point")
            print("Start and Stop will be exchanged")
            self.start, self.stop = self.stop, self.start
    
    """Processing functionality"""
    def prepare_profile_coordinates(self):
        """Prepare the evaluation coordinates as stack"""
        length = self.length
        x0, y0 = self.start     
        x1, y1 = self.stop
        x = np.linspace(x0, x1, length+1)
        y = np.linspace(y0, y1, length+1)
        self.profile_coords = np.vstack((y,x))
    
    @property
    def normal_vector(self):
        """Normal vector of line"""
        return self._det_normal_vec()
        
    @property
    def length(self):
        """Determine the length in grid coordinates"""
        del_x, del_y = self._delx_dely()
        return int(round(np.hypot(del_x, del_y)))
    
    def get_line_profile(self, array, **kwargs):
        """Retrieve line profile of data on a 2D array
        
        Uses method :func:`scipy.ndimage.map_coordinates` 
        
        Note
        ----
        The spline interpolation will fail if there are NaNs in the input data. 
        In this case, try using order=1 as additional input argument or use
        prefilter=False. For details see `here <https://docs.scipy.org/doc/
        scipy-0.14.0/reference/generated/scipy.ndimage.interpolation.
        map_coordinates.html>`__
        
        Parameters
        ----------
        array : ndarray
            2D data array
        **kwargs
            additional keyword args that are passed to
            :func:`scipy.ndimage.map_coordinates`
            
        Returns
        -------
        ndarray
            1D array containing the values along the line profile coordinates
        """
        # Extract the values along the line, using cubic interpolation
        zi = map_coordinates(array, self.profile_coords, **kwargs)
        return zi
    
    def plot_line_on_grid(self, array, ax=None, **kwargs):
        """Plot this line on an input array using imshow
        
        :param ndarray array: the data array
        :param **kwargs: additional keyword arguments for imshow
        
        """
        import matplotlib.pyplot as plt
        if ax is None:
            fig, ax = plt.subplots(1,1)
        
        im = ax.imshow(array, cmap="gray", interpolation='none', 
                       origin="upper", **kwargs)
        fig.colorbar(im, ax=ax, shrink=0.9)
        ax.plot([self.start[0], self.stop[0]], [self.start[1], self.stop[1]],
                'co-')
            
        ax.set_xlim([0, array.shape[1] - 1])
        ax.set_ylim([array.shape[0] - 1, 0])
        return ax
    
    def plot_line_profile(self, array, ax=None, **kwargs):
        """Plot the line profile 
        
        :param ndarray array: the data array
        :param ax (None): axes object
        :param **kwargs: additional keyword arguments for imshow
        
        """
        if ax is None:
            import matplotlib.pyplot as plt
            fig, ax = plt.subplots(1,1)
        p = self.get_line_profile(array)
        ax.set_xlim([0, self.length])
        ax.grid()
        ax.plot(p)
        ax.set_title("Line profile")
        return ax
    
    def plot(self, array, fig_width=12):
        """High level plot function, calls:
        
            1. :func:`plot_line_on_grid`
            #. :func:`plot_line_profile`
            
        and puts them next to each other in a subplot
        
        :param ndarray array: the data array
        :param int fig_width: width of figure in inch
        
        """
        import matplotlib.pyplot as plt

        dx, dy = self._delx_dely()
        if dx > dy:
            r = float(dy) / dx
            h = int(fig_width * r * 2.5)
            fig, axes = plt.subplots(2,1, figsize=(fig_width, h))
        else:
            fig, axes = plt.subplots(1,2, figsize = (18, 6))
        self.plot_line_on_grid(array, ax=axes[0])
        self.plot_line_profile(array, ax=axes[1])
        plt.tight_layout()
    
    """'Private' functions"""
    def _delx_dely(self):
        """Returns length of x and y range"""
        return (abs(self.stop[0] - self.start[0]),
                abs(self.stop[1] - self.start[1]))
                        
    def _det_normal_vec(self):
        """Determines the normal vector of the line"""
        delx, dely = self._delx_dely()
        try:
            a = np.arctan(float(dely) / float(delx))
        except ZeroDivisionError:
            a = np.pi / 2
        return (np.sin(a), np.cos(a))
        
    """Magic methods"""
    def __str__(self):
        """String representation"""
        s = ("LineOnGrid %s\nStart (x,y): %s\nStop (x,y): %s\n" 
                                        %(self.id, self.start, self.stop))
        return s


# -*- coding: utf-8 -*-
"""Test environment for base.py module."""
from __future__ import (absolute_import, division)
import numpy.testing as npt
from geonum.test.markers import srtm_avail, netcdf4_avail
import geonum.topodataaccess as tp

@netcdf4_avail
def test_etopo1_init():
    acc = tp.Etopo1Access(check_access=False, search_database=False)
    from geonum import LOCAL_TOPO_PATH
    assert LOCAL_TOPO_PATH == acc.local_path
    assert acc.file_name == 'ETOPO1_Ice_g_gmt4.grd'
    
@srtm_avail
def test_srtm():
    acc = tp.SRTMAccess()
    d = acc.get_data(acc._TESTLAT, acc._TESTLON)
    npt.assert_equal(d.data.shape, (2, 2))
    vals = [d.data.std(), d.data.mean(), d.data.min(), d.data.max()]
    
    npt.assert_array_almost_equal(vals, [4.0, 857.0, 853.0, 861.0])

def test_topoaccess_invalid_mode():
    all_ok = True
    try:
        tp.TopoDataAccess('bla')
    except tp.InvalidTopoMode:
        pass
    except:
        all_ok = False
    
    acc = tp.TopoDataAccess()
    try:
        acc.get_data(0,0, mode='bla')
    except tp.InvalidTopoMode:
        pass
    except:
        all_ok = False    
    assert all_ok
    
if __name__=='__main__':
    test_etopo1_init()
    test_srtm()
    test_topoaccess_invalid_mode()


# -*- coding: utf-8 -*-
#
# Geonum is a Python library for geographical calculations in 3D
# Copyright (C) 2017 Jonas Gliss (jonasgliss@gmail.com)
#
# This program is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License a
# published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""
Helper methods for geonum library
"""
import os
import numpy as np
import matplotlib.cm as colormaps
import matplotlib.colors as colors
from matplotlib.pyplot import draw

exponent = lambda num: np.int(np.floor(np.log10(np.abs(num))))

def all_topodata_search_dirs():
    """Returns a list of all directories that are searched for topography data
    
    Returns
    -------
    list
        list containing all search directories (absolute paths)
    """
    from geonum import TOPO_INFO_FILE, LOCAL_TOPO_PATH
    paths = [LOCAL_TOPO_PATH]
    with open(TOPO_INFO_FILE, "r") as f:
        for line in f:
            p = line.strip()
            if os.path.exists(p) and not p in paths:
                paths.append(os.path.normpath(p))
    return paths
    
def check_and_add_topodir(local_dir):
    """Check if input directory is registered and if not, register it
    
    Parameters
    ----------
    local_dir : str
        directory that is supposed to be checked
    
    """
    p = os.path.normpath(local_dir)
    if not p in all_topodata_search_dirs():
        from geonum import TOPO_INFO_FILE
        with open(TOPO_INFO_FILE, "a") as f:
           f.write('{}'.format(p))
           print('Adding new default local topo data path to '
                 'file LOCAL_TOPO_DATA.txt:\n{}'.format(p))
           
def isnum(val):
    """Checks if input is number (int or float) or and not nan
    
    :returns: bool, True or False    
    """
    if isinstance(val, (int, float)) and not np.isnan(val):
        return True
    return False

def rotate_xtick_labels(ax, deg=30, ha="right"):
    """Rotate xtick labels in matplotlib axes object"""
    draw()
    lbls = ax.get_xticklabels()
    lbls = [lbl.get_text() for lbl in lbls]
    ax.set_xticklabels(lbls, rotation = 30, ha = "right")
    draw()
    return ax
    
def haversine_formula(lon0, lat0, lon1, lat1, radius=6371.0):
    """Haversine formula
    
    Approximate horizontal distance between 2 points assuming a spherical 
    earth
    
    :param float lon0: longitude of first point in decimal degrees
    :param float lat0: latitude of first point in decimal degrees
    :param float lon1: longitude of second point in decimal degrees
    :param float lat1: latitude of second point in decimal degrees
    :param float radius (6371.0): average earth radius in km
    """
    hav = lambda d_theta: np.sin(d_theta / 2.0) ** 2
    
    d_lon = np.radians(lon1 - lon0)
    d_lat = np.radians(lat1 - lat0)
    lat0 = np.radians(lat0)
    lat1 = np.radians(lat1)
 
    a = hav(d_lat) + np.cos(lat0) * np.cos(lat1) * hav(d_lon)
    c = 2 * np.arcsin(np.sqrt(a))
 
    return radius * c

def approximate_connection_vector(lon0, lat0, lon1, lat1, len_lat_km=111.20):
    """Returns approximate connection vector between two points
    
    :param float lon0: longitude of first point in decimal degrees
    :param float lat0: latitude of first point in decimal degrees
    :param float lon1: longitude of second point in decimal degrees
    :param float lat1: latitude of second point in decimal degrees
    
    Careful: Only approximative, suited for small distance ( < 100km)
    """    
    
        
    lat = (lat0 + lat1) / 2 * 0.01745
    dx = len_lat_km * np.cos(lat) * (lon1 - lon0)
    dy = len_lat_km * (lat1 - lat0)
    return np.array((dx,dy))
    
def shifted_color_map(vmin, vmax, cmap=None):
    """Shift center of a diverging colormap to value 0
    
    .. note::
    
        This method was found `here <http://stackoverflow.com/questions/
        7404116/defining-the-midpoint-of-a-colormap-in-matplotlib>`_ 
        (last access: 17/01/2017). Thanks to `Paul H <http://stackoverflow.com/
        users/1552748/paul-h>`_ who provided it.
    
    Function to offset the "center" of a colormap. Useful for
    data with a negative min and positive max and if you want the
    middle of the colormap's dynamic range to be at zero level
    
    :param vmin: lower end of data value range
    :param vmax: upper end of data value range
    :param cmap: colormap (if None, use default cmap: seismic)
    
    :return: 
        - shifted colormap
        
    """
    #midpoint = 1 - np.abs(im.max())/(np.abs(im.max()) + np.abs(im.min()))
    if cmap is None:
        cmap = colormaps.seismic
        
    midpoint = 1 - np.abs(vmax)/(np.abs(vmax) + np.abs(vmin))
    
    cdict = {
        'red': [],
        'green': [],
        'blue': [],
        'alpha': []
    }

    # regular index to compute the colors
    reg_index = np.linspace(0, 1, 257)

    # shifted index to match the data
    shift_index = np.hstack([
        np.linspace(0.0, midpoint, 128, endpoint=False), 
        np.linspace(midpoint, 1.0, 129, endpoint=True)
    ])

    for ri, si in zip(reg_index, shift_index):
        r, g, b, a = cmap(ri)

        cdict['red'].append((si, r, r))
        cdict['green'].append((si, g, g))
        cdict['blue'].append((si, b, b))
        cdict['alpha'].append((si, a, a))

    #newcmap = colors.LinearSegmentedColormap('shiftedcmap', cdict)
    #register_cmap(cmap=newcmap)

    return colors.LinearSegmentedColormap('shiftedcmap', cdict)


# -*- coding: utf-8 -*-
#
# Geonum is a Python library for geographical calculations in 3D
# Copyright (C) 2017 Jonas Gliss (jonasgliss@gmail.com)
#
# This program is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License a
# published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from LatLon23 import GeoVector
from geonum.geopoint import GeoPoint
    
from numpy import (radians, cos, sin, degrees, sqrt,tan, isnan, arctan2)

class GeoVector3D(GeoVector):
    """A 3 dimensional geo vector object 
    
    3D vector representation for geo numerical calculations, intuitive 
    usage, i.e.::
    
        from geonum import GeoPoint, GeoVector3D
        p = GeoPoint(10.0, 15.0, name = "random_point") #lat, lon
        v = GeoVector3D(dx = 15, dy = 100, dz = 300) #dx, dy in km, dz in m
        
        new_point = p + v # GeoPoint object
        
    
    Note
    ----
    
        This class inherits and makes use of the functionality of 
        `GeoVector` objects of the `LatLon` library. Methods and 
        attributes are partly the same and partly overwritten
        (note that it is not initiated as :class:`GeoVector`)
 
    """
            
    def __init__(self, dx=None, dy=None, dz=None, azimuth=None, dist_hor=None, 
                 elevation=None, anchor=None, name="n/d"):
        """Class initialisation
        
        :param float dx: longitudinal length in km
        :param float dy: latitudinal length in km
        :param float dz: altitude length in m
        :param float azimuth: azimuth angle to North direction (corresponds to 
            heading_initial variable in :class:`GeoVector` object)
        :param float dist_hor: horizontal displacement of object in km
        :param float elevation: elevation angle in decimal degrees 
            (0 corresponds to horizon, 90, to zenith)
        
        For initiation use one of the following to input combinations:
        
            1. dx, dy, dz
            #. dz, azimuth, dist_hor
            #. azimuth, dist_hor, elevation
            
        Multiple input combinations will be processed in the preferred order
        as given in the list
        
        """
        self.name = name
        if any(x == None for x in [dx, dy]): # If only initial_heading and distance are given
            theta_rad = radians(self._angle_or_heading(azimuth))
            self.dx = dist_hor * cos(theta_rad)
            self.dy = dist_hor * sin(theta_rad)
        elif azimuth == None and dist_hor == None: # If only dx and dy are given
            self.dx = dx
            self.dy = dy
        else:
            raise NameError
        #Check input for altitude difference dz
        if dz is None or isnan(dz): #invalid for dz directly
            if elevation is not None and -90 <= elevation <= 90: #check if instead elevation is valid, then set dz
                #tan elev = dz/dist_hor
                dz = tan(radians(elevation))*sqrt(self.dx**2+self.dy**2)*1000
            else: #both dz input and elevation are invalid, set dz=0
                dz = 0.0
        self.dz = dz
        # dictionary with private attributes
        self._priv_attr= {"anchor" : None}
        # call setter for private attribute anchor (this ensures that input 
        # attr anchor is of right type
        try:
            self.set_anchor(anchor)
        except:
            pass
    
    @property
    def azimuth(self):
        """Azimuth angle in dec degrees 
        
        Horizontal orientation measured from N direction
        """
        return self._geom_hor()[0]

    @property
    def elevation(self):
        """Elevation angle in decimal degrees
        
        Measured from horizon (i.e. 0 -> horizon, 90 -> zenith)
        """
        return 90 - self.polar_angle    
    
    @property
    def polar_angle(self):
        """Polar angle in decimal degrees
        
        Returns the polar angle in decimal degrees (measured from zenith), 
        i.e.::
        
            90 - self.elevation
            
        """
        return degrees(arctan2(self.dist_hor, self.dz / 1000.))
    
    @property    
    def magnitude(self):
        """Return magnitude of vector (length)"""
        return sqrt(self.dist_hor**2 + (self.dz / 1000.)**2)

    @property
    def norm(self):
        """Return magnitude of vector (length)"""
        return self.magnitude
        
    @property
    def dist_hor(self):
        """Horizontal distance spanned by this vector"""
        return self._geom_hor()[1]
    
    @property
    def anchor(self):
        """Getter method for private attribute anchor"""
        return self._priv_attr["anchor"]
        
    @anchor.setter
    def anchor(self, value):
        """Setter method of private property anchor
        
        See :func:`set_anchor` for input / output specs
        """     
        self.set_anchor(value)
    
    def set_anchor(self, geo_point):
        """Set anchor of this vector
        
        :param GeoPoint geo_point: anchor point of this vector
        :raises TypeError: if input is not :class:`GeoPoint` object        
        """
        if not isinstance(geo_point, GeoPoint):
            raise TypeError("Could not set anchor: Invalid input type")
        self._priv_attr["anchor"] = geo_point
           
    def intersect_hor(self, other):
        """Determine the horizontal intersection of this vector with other 
        input vector
        
        :param GeoVector3D: other vector
        
        .. note::
        
            Only works if anchor is defined
        """
        if not all([isinstance(x,GeoPoint) for x in [self.anchor, other.anchor]]):
            raise ValueError("Intersection can not be determined, anchor "
                " of one of the vectors not set..")
        v = other.anchor - self.anchor
        v.dz = 0.0
        other_az = radians((other.azimuth + 360) % 360)
        self_az = radians((self.azimuth + 360) % 360)
        dy = (v.dx - tan(other_az) * v.dy) / (tan(self_az) - tan(other_az))
        dx = tan(self_az) * dy
        return GeoVector3D(dx, dy, dz=0.0)
        
    def _geom_hor(self):
        """Returns horizontal heading and horizontal magnitude"""
        return (degrees(arctan2(self.dx, self.dy)),
                sqrt(self.dx**2 + self.dy**2))
        
    """Plotting etc..."""
    def plot(self, map, add_anchor=False, **kwargs):
        """Plot this vector into existing basemap
        
        :param Map map: map object
        :param bool add_anchor: If true, the anchor point is plotted as well
        :param kwargs: additional keyword arguments
        
        .. note::
        
            1. The basemap needs to be set up with Axes3D
            #. ``self.anchor`` must be set with a :class:`Geopoint` instance
            
        """
        map.draw_geo_vector_3d(self, **kwargs)
        if add_anchor:
            self.anchor.plot(map, add_name=True, dz_text=self.dz*.1)
        
    """Redifining magic methods from base class :class:`GeoVector` object for
    3D calcs
    """
    def __call__(self):
        """Call function
        
        :returns:
            - float azimuth angle
            - float, horizontal length in km
            - float, vertical length in m
        """
        azimuth, dist_hor = self._geom_hor()
        return azimuth, dist_hor, self.dz
    
    
        
    def __neg__(self):
        """Returns negative of this vector"""
        return GeoVector3D(-self.dx, -self.dy, -self.dz)
    
    def __add__(self, other):
        """Add another geo vector
        
        
        :param GeoVector3D other: another geo vector  
        """
        return GeoVector3D(self.dx + other.dx, self.dy + other.dy,
                           self.dz + other.dz)

    def __str__(self):
        """String representation"""
        return ('GeoVector3D {}\n'
                'Azimuth: {:.2f}°, Elevation: {:.4f}°, Magnitude: {:.2f} km '
                '(hor: {:.2f} km)'
                .format(self.name, self.azimuth, self.elevation, 
                        self.magnitude, self.dist_hor))
            
    
    def __repr__(self):
        return ('Az %s, Elev %s, Mag. %s' 
                %(self.azimuth, self.elevation, self.magnitude))
    
    def type(self):
        """Returns object type"""
        return 'GeoVector3D'



# -*- coding: utf-8 -*-
"""Test environment for base.py module."""
from __future__ import (absolute_import, division)
import pytest
from geonum import SRTM_AVAILABLE, NETCDF_AVAILABLE

# custom skipif marker that is used below for test functions that 
# require SRTM.py to be installed
srtm_avail = pytest.mark.skipif(SRTM_AVAILABLE==False,
                   reason='Skipping SRTM database tests. srtm.py library is '
                   'not installed')

netcdf4_avail = pytest.mark.skipif(NETCDF_AVAILABLE==False,
                   reason='Skipping tests requiring NetCDF4 library.')
                    #allow_module_level=True)